/****************************************************************************
 * filter.c
 *
 * Copyright (c) 2000 Stefan Hellkvist, Ericsson Radio Systems AB
 * Email: stefan@hellkvist.org
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *****************************************************************************/

/* Oct 2000 - ported to Pocket Viewer by Wittawat Yamwong <wittawat@web.de> */

/*ce Comment: all filters allocate a new pattern - the old one is preserved */ 

#include "modeldef.h"

#include "md.h"

#include "filter.h"
#include "pattern.h"
#include "point.h"



/*
#define NORM_WIDTH  100
#define NORM_HEIGHT 100
*/
#define NORM_SIZE   100
#define THRESHOLD   36

#define MAX( x, y ) max(x,y)

/*
        newx = (x * a - xmin*a)/b
        newy = (x * a - ymin*a)/b

        a = NORM_SIZE-1
        b = max(dx,dy)
*/
Pattern FAR *
sizeNormFilter( Pattern FAR *p )
{
    int i, xmin, xmax, ymin, ymax;
    int delta;
    Point tmp, po;
    Pattern FAR *np;

    if ( p == NULL )
        return NULL;
    np = newPattern( p->size, p->ascii );
    if ( np == NULL )
        return NULL;

    tmp = p->pv[ 0 ];
    xmin = xmax = tmp.x;
    ymin = ymax = tmp.y;

    for ( i = 1; i < p->nPoints; i++ )
    {
        tmp = p->pv[ i ];
        if ( tmp.x < xmin )
            xmin = tmp.x;
        if ( tmp.y < ymin )
            ymin = tmp.y;
        if ( tmp.x > xmax )
            xmax = tmp.x;
        if ( tmp.y > ymax )
            ymax = tmp.y;
    }

    delta = MAX(xmax - xmin, ymax - ymin);
    xmin *= NORM_SIZE-1;
    ymin *= NORM_SIZE-1;

    for ( i = 0; i < p->nPoints; i++ )
    {
        tmp = p->pv[ i ];
        po.x = ((NORM_SIZE-1) * tmp.x - xmin) / delta;
        po.y = ((NORM_SIZE-1) * tmp.y - ymin) / delta;
        addPoint( np, po );
    }

    return np;
}


static
int
distanceSq( Point p0, Point p1 )
{
    return (( p0.x - p1.x ) * ( p0.x - p1.x ))
        + (( p0.y - p1.y ) * ( p0.y - p1.y ));
}

Pattern FAR *
pointDistanceNormFilter( Pattern FAR *p )
{
    Pattern FAR *np;
    Point last, tmp;
    int i;

    if ( p == NULL )
        return NULL;
    np = newPattern( p->size, p->ascii );
    if ( np == NULL )
        return NULL;

    last = p->pv[ 0 ];
    addPoint( np, last );
    for ( i = 1; i < p->nPoints; i++ )
    {
        tmp = p->pv[ i ];
        if ( distanceSq( tmp, last ) >= THRESHOLD )
        {
            last = tmp;
            addPoint( np, last );
        }
    }

    return np;
}


Pattern FAR *
smoothFilter( Pattern FAR *p )
{
    Pattern FAR *np;
    Point p1, p2, p3, tmp;
    int i;

    if ( p == NULL )
        return NULL;
    np = newPattern( p->size, p->ascii );
    if ( np == NULL )
        return NULL;

    addPoint( np, p->pv[ 0 ] );
    for ( i = 1; i < p->nPoints - 1; i++ )
    {
        p1 = p->pv[ i - 1 ];
        p2 = p->pv[ i ];
        p3 = p->pv[ i + 1 ];
        tmp.x = ( p1.x + p2.x + p3.x ) / 3;
        tmp.y = ( p1.y + p2.y + p3.y ) / 3;
        addPoint( np, tmp );
    }
    addPoint( np, p->pv[ p->nPoints - 1 ] );
    return np;
}


/* correct a slanted input pattern */

static int slantedFilterStrength = 9; 
/* >8=do nothing, 
    8=minimum correction, ..., 
    2=correct about l5 degree, 
    1=correct about 26 degree, 
    0=correct about 45 degree */ 

Pattern FAR *antiSlantedFilter( Pattern FAR *p )
{
    Pattern FAR *np;
    Point tmp;
    int i, ymin;

    if ( p == NULL )
        return NULL;
    np = newPattern( p->size, p->ascii );
    if ( np == NULL )
        return NULL;

    ymin = p->pv[0].y;
    if (slantedFilterStrength<=8)
        for ( i = 0; i < p->nPoints; i++ )
            if ( p->pv[0].y < ymin )
                ymin = p->pv[0].y;
            
    for ( i = 0; i < p->nPoints; i++ )
    {
        tmp = p->pv[ i ];
        if (slantedFilterStrength<=8)
            tmp.x += ((tmp.y-ymin)>>slantedFilterStrength);
        addPoint( np, tmp );
    }
    return np;
}

void setSlantedFilterStrength(int s)
{
    slantedFilterStrength = s;
}
