/*
 *  Memory Manager
 *  Copyright (C) 2000 Wittawat Yamwong <wittawat@web.de>
 */
#include "mm.h"

static u8 FAR *heap;
static u8 FAR *freelist;

#define ALIGN 8
#define SIZE(p) (((freenode_t FAR *)(p))->size)
#define NEXT(p) (((freenode_t FAR *)(p))->next + heap)
#define PREV(p) (((freenode_t FAR *)(p))->prev + heap)
#define HDR_LEN 2
#define SET_NEXT(p,value) ((freenode_t FAR *)(p))->next = (value) - heap
#define SET_PREV(p,value) ((freenode_t FAR *)(p))->prev = (value) - heap
#define COPY_NODE(a,b) *(freenode_t FAR *)(b) = *(freenode_t FAR *)(a)

typedef struct {
    u16 size;
    u16 next;
    u16 prev;
} freenode_t;

static void remove_from_freelist(u8 FAR *p, u16 size);
static void add_to_freelist(u8 FAR *p, u16 size);

void FAR * mm_malloc(u16 size)
{
    u8 FAR *p;
    if (size == 0)
	return 0;
    size = ((size + 2*sizeof(u16) + ALIGN-1) & ~(ALIGN-1));
    p = NEXT(freelist);
    while (p != freelist && SIZE(p) < size)  p = NEXT(p);
    if (p != freelist) {
	remove_from_freelist(p,size);
	SIZE(p+size-2) = SIZE(p) = size | 1;
	p += HDR_LEN;
    } else
	p = 0;
    return p;
}

void mm_free(void FAR *p__)
{
    u16 size,temp;
    u8 FAR *p;

    if (p__ == 0)
	return;
    p = (u8 FAR *)p__ - HDR_LEN;
    size = SIZE(p);
    if ((size & 1) == 0)
	return; /* It's already free! */
    size &= ~1; /* clear alloc flag */
    temp = SIZE(p + size); /* this requires guard block at the end */
    if ((temp & 1) == 0) { /* is the next block free? */
	remove_from_freelist(p+size,temp);
	size += temp;
    }
    temp = *(u16 FAR *)(p - 2); /* require guard block at the begining */
    if ((temp & 1) == 0) {
	p -= temp;
	remove_from_freelist(p,temp);
	size += temp;
    }
    add_to_freelist(p, size);
}

void mm_create_heap(void FAR *h, u16 size)
{
    size = size & ~(ALIGN-1);
    heap = (u8 FAR *) h;
    freelist = heap + size - ALIGN;
    *(u16 FAR *)(heap + ALIGN - 2) = ALIGN | 1; /* begin guard */
    SIZE(freelist) = ALIGN | 1;		        /* end guard */
    SET_NEXT(freelist,freelist);
    SET_PREV(freelist,freelist);
    add_to_freelist(heap + ALIGN,size - 2*ALIGN);
}

static void remove_from_freelist(u8 FAR *p, u16 size)
{
    if (SIZE(p) == size) {
	SET_NEXT(PREV(p),NEXT(p));
	SET_PREV(NEXT(p),PREV(p));
    } else {
	COPY_NODE(p,p+size);
	p += size;
	SET_NEXT(PREV(p),p);
	SET_PREV(NEXT(p),p);
	SIZE(p) -= size;
	*(u16 FAR *)(p + SIZE(p) - 2) = SIZE(p);
    }
}

static void add_to_freelist(u8 FAR *p, u16 size)
{
    SIZE(p) = size;
    *(u16 FAR *)(p+size-2) = size;
    SET_PREV(NEXT(freelist),p);
    SET_NEXT(p,NEXT(freelist));
    SET_PREV(p,freelist);
    SET_NEXT(freelist,p);
}


#ifdef TEST_MM
#include <stdio.h>
#include <stdlib.h>

int main()
{
    const int heap_size = 10*1024;
    void *mem = mm_malloc(heap_size);
    int size;
    void *p;
    char buf[50];

    mm_create_heap(mem,heap_size);
    printf("size = %d\n",heap_size);
    do {
	printf("Quit,Alloc,Free,Trace: ");
	scanf("%s",buf);
	switch (buf[0]) {
	case 'q':
	    exit(0);

	case 'a':
	    printf("size = ");
	    scanf("%d",&size);
	    p = mm_malloc(size);
	    printf("mm_alloc(%d) = %d\n",size,(u8 *)p - heap);
	    break;

	case 'f':
	    printf("ptr = ");
	    scanf("%d",&size);
	    mm_free(size + heap);
	    break;

	case 't':
	    printf("freelist dump\n");
	    p = NEXT(freelist);
	    while (p != freelist) {
		printf("  ofs %d, size %d\n",(u8*)p-heap,SIZE(p));
		p = NEXT(p);
	    }
	    break;
	}
	printf("\n");
    } while (1);
}
#endif TEST_MM
