/*********************************************************
 * modebut.c
 *
 * Copyright (c) Christian Ensel
 *
 * Merlin-Mode-Buttons (and indicators at the same time)
 *********************************************************/

#include "modeldef.h"

#include "md.h"

#include "util.h"
#include "modebut.h"

static TCHTBL mbtch[NRMODEBUTTONS+1];
static T_ICON mbicon[NRMODEBUTTONS];

/* button-types */
typedef enum { ARRAYEND = -1,  /* marker for end of array */
               STATE_ALPHA     = ALPHA_MODE        /* lower case */ ,
               STATE_SHIFTED   = 0xF0 + ALPHA_MODE /* upper case, shifted */ ,
               STATE_CAPSLOCK  = 0xF1 + ALPHA_MODE /* upper case locked */ ,
               STATE_NUM       = NUMERAL_MODE      /* numbers */ ,
               STATE_NUM_EXT   = NUMERAL_EXT_MODE  /* ops and brackets */ ,
               STATE_ALPHA_EXT = ALPHA_EXT_MODE    /* extendend chars */ } t_butstate;

/* possible button-changes for each button, if it is clicked several times */
t_butstate but0_statetrans[] = { STATE_ALPHA, STATE_SHIFTED, STATE_CAPSLOCK, ARRAYEND };
t_butstate but1_statetrans[] = { STATE_ALPHA_EXT, ARRAYEND };
t_butstate but2_statetrans[] = { STATE_NUM, STATE_NUM_EXT, ARRAYEND };

t_butstate *butstatetrans[NRMODEBUTTONS] = { but0_statetrans, but1_statetrans, but2_statetrans };

/* current button and highlighting states */
int activebut;  /* -1 for none */
static int butstate_index[NRMODEBUTTONS];  

#define CURRENTSTATE (butstatetrans[activebut][butstate_index[activebut]])


/* invert a Button */
void invButton(int n)
{
    LibReverse( mbtch[n].x1, mbtch[n].y1, MODEBUTTON_XSIZE, MODEBUTTON_YSIZE );
    LibReverse( mbtch[n].x1+1, mbtch[n].y1+1, MODEBUTTON_XSIZE-3, MODEBUTTON_YSIZE-3 );
    LibPutDispBox(mbtch[n].x1, mbtch[n].y1, MODEBUTTON_XSIZE, MODEBUTTON_YSIZE);
}

void modeButtonDraw(int but)
{
    int basex, basey;
    LibClrBox( mbtch[but].x1, mbtch[but].y1, MODEBUTTON_XSIZE, MODEBUTTON_YSIZE );
    basex = mbtch[but].x1+1;
    basey = mbtch[but].y1+2;
    switch ( (butstatetrans[but][butstate_index[but]]) ) {
        case STATE_ALPHA:
            LibPutProFont(IB_PFONT2, 'a', basex+1, basey);
        break;
        case STATE_SHIFTED:
            LibPutProFont(IB_PFONT1, 'A', basex+1, basey);
        break;
        case STATE_CAPSLOCK:
            LibPutProFont(IB_PFONT2, 'A', basex, basey);
        break;
        case STATE_NUM:
            LibPutProFont(IB_PFONT2, '1', basex+1, basey);
        break;
        case STATE_NUM_EXT:
            LibPutProFont(IB_PFONT1, '+', basex+1, basey);
        break;
        case STATE_ALPHA_EXT:
            LibPutProFont(IB_PFONT1, ';', basex+3, basey);
        break;
        LibIconPrint(&mbicon[but]);
    }
}

#define AUTOSELECTSTATE -1
void activateButton(int n, int stateindex )
/* n:          buttonnr, 
   stateindex: the index to the state that this button should have;
               use AUTOSELECTSTATE for default behaviour after click on this button */
{
    if ( activebut != n ) {  /* new button chosen */
        /* deselect prev. button */
        if ( activebut >= 0 )
            invButton(activebut);
        activebut = n;
        if ( stateindex == AUTOSELECTSTATE
             || butstate_index[activebut] == stateindex ) { 
            /* wanted button already displayed */
            invButton(activebut);
            return;
        }      
    }          
    else if ( butstate_index[activebut] == stateindex )  /* already active (and highlighted) */
        return;
               
    if ( stateindex == AUTOSELECTSTATE ) {
        butstate_index[activebut]++;
        if ( CURRENTSTATE == ARRAYEND )  /* circle index around */
            butstate_index[activebut] = 0;
    } else
        butstate_index[activebut] = stateindex;
        
    /* butstate changed -> redraw button */    
    modeButtonDraw(activebut);
    invButton(activebut);
}


/* refresh displays status after a mode change */
void modeButtonsDisplayMode() /*  t_char_mode char_mode, t_shift_mode shift_mode) */
{
    int i;
    
    /* map Char/Shift-Mode to but-state */
    t_butstate bs = (t_butstate)getCharMode();  
    if ( bs == STATE_ALPHA ) {
        t_shift_mode sm = getShiftMode();
        if ( sm == SHIFTED_MODE )
            bs = STATE_SHIFTED;
        else if ( sm == CAPSLOCK_MODE )
            bs = STATE_CAPSLOCK;
    }
    if ( activebut != -1 && CURRENTSTATE == bs )
        return;
        
    /* see whether one of the currently displayed buttons already in in this state */
    for (i=0; i<NRMODEBUTTONS; i++) 
        if ( bs == butstatetrans[i][butstate_index[i]] ) {
            activateButton(i,AUTOSELECTSTATE);
            break;
        }           
    if ( i==NRMODEBUTTONS )  { /* find button that is able to express the current mode  */
        int j;
        for (i=0; i<NRMODEBUTTONS; i++) {
            t_butstate testbs;
            j=0;
            while ( (testbs = butstatetrans[i][j]) != ARRAYEND && testbs != bs )
                j++;
            if ( testbs == bs )
                break;
        }
        activateButton(i,j);
    }
}

/* check pen-events and change mode if a mode-button was pressend */
void modeButtonsPenEvent(TCHSTS *tsts)
{
    int buttonnr = tsts->obj - MODEBUTTON_OBJ_BASE;
    if ( buttonnr < 0 || buttonnr >= NRMODEBUTTONS )
        return;

    if ( LibIconClick( (T_ICON FAR *)(&mbicon[buttonnr]), tsts ) == TRUE )  {
        /* a button was pressed */
        activateButton(buttonnr,AUTOSELECTSTATE);
        switch ( CURRENTSTATE ) {
            case STATE_ALPHA:
                setCharMode( ALPHA_MODE );
                setShiftMode( LOWERCASE_MODE );
            break;
            case STATE_SHIFTED:
                setCharMode( ALPHA_MODE );
                setShiftMode( SHIFTED_MODE );
            break;
            case STATE_CAPSLOCK:
                setCharMode( ALPHA_MODE );
                setShiftMode( CAPSLOCK_MODE );
            break;
            case STATE_NUM:
                setCharMode( NUMERAL_MODE );        
            break;
            case STATE_NUM_EXT:
                setCharMode( NUMERAL_EXT_MODE );
            break;
            case STATE_ALPHA_EXT:
                setCharMode( ALPHA_EXT_MODE );
            break;
        }
        
    }

}
                      
void modeButtonsInit( s_modeButtonsPos* mbPos )
{
    int i;
    /* set TCHTBL */
    for ( i=0; i<NRMODEBUTTONS; i++) {
        mbtch[i].x1 = mbPos->x[i]; 
        mbtch[i].y1 = mbPos->y[i]; 
        mbtch[i].x2 = mbPos->x[i] + MODEBUTTON_XSIZE; 
        mbtch[i].y2 = mbPos->y[i] + MODEBUTTON_YSIZE; 
        mbtch[i].act = ACT_ICON;
        mbtch[i].obj = MODEBUTTON_OBJ_BASE + i;
        mbtch[i].ext = 0x0000;
    }
    mbtch[NRMODEBUTTONS] = TCHTBL_END;
    
    /* set and display ICONs */
    for ( i=0; i<NRMODEBUTTONS; i++) {
        mbicon[i].tch = &mbtch[i]; 
        mbicon[i].ngp = NULL; 
        mbicon[i].rgp = NULL; 
        mbicon[i].kind = 0x03; /**/
  	    butstate_index[i] = 0;
        modeButtonDraw(i);
    }
    activebut=-1;    
    modeButtonsDisplayMode();
    /*LibPutDisp();*/

  	LibTchStackPush( (TCHTBL FAR *)(&mbtch) );
}



